package dev.peerat.parser.java.printer;

import java.util.Iterator;

import dev.peerat.parser.java.JavaElement;
import dev.peerat.parser.java.value.ArrayAccessValue;
import dev.peerat.parser.java.value.ArrayValue;
import dev.peerat.parser.java.value.BiValue;
import dev.peerat.parser.java.value.CastValue;
import dev.peerat.parser.java.value.InstanceValue;
import dev.peerat.parser.java.value.LambdaValue;
import dev.peerat.parser.java.value.LambdaValue.LambdaParameter;
import dev.peerat.parser.java.value.MethodCallValue;
import dev.peerat.parser.java.value.ModifierValue;
import dev.peerat.parser.java.value.StaticValue;
import dev.peerat.parser.java.value.TriValue;
import dev.peerat.parser.java.value.Value;
import dev.peerat.parser.java.value.VariableAccessValue;

public class JavaValuePrinter{
	
	public static class JavaStaticValuePrinter extends JavaPrinter<StaticValue>{
		
		public void print(StaticValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception {
			writer.write(buffer);
			writer.write(element.getToken());
		}
		
	}
	
	public static class JavaBiValuePrinter extends JavaPrinter<BiValue>{
		
		public void print(BiValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer);
			Value left = element.left();
			provider.getPrinter(left).print(left, writer, "", provider);
			
			writer.write(element.getAction());
			
			Value right = element.right();
			provider.getPrinter(right).print(right, writer, "", provider);
		}
		
	}
	
	public static class JavaTriValuePrinter extends JavaPrinter<TriValue>{
		
		public void print(TriValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer);
			Value value = element.getChecker();
			provider.getPrinter(value).print(value, writer, "", provider);
			
			writer.write(" ? ");
			
			value = element.success();
			provider.getPrinter(value).print(value, writer, "", provider);
			
			writer.write(" : ");
			
			value = element.fail();
			provider.getPrinter(value).print(value, writer, "", provider);
		}
		
	}
	
	public static class JavaArrayAccessValuePrinter extends JavaPrinter<ArrayAccessValue>{
		
		public void print(ArrayAccessValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer);
			Value value = element.base();
			provider.getPrinter(value).print(value, writer, "", provider);
			writer.write("[");
			value = element.getAccessor();
			provider.getPrinter(value).print(value, writer, "", provider);
			writer.write("]");
		}
		
	}
	
	public static class JavaArrayValuePrinter extends JavaPrinter<ArrayValue>{
		
		public void print(ArrayValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer+"{");
			Value[] values = element.getValues();
			for(int i = 0; i < values.length; i++){
				Value value = values[i];
				provider.getPrinter(value).print(value, writer, "", provider);
				if(i < values.length-1) writer.write(", ");
			}
			writer.write("}");
		}
		
	}
	
	public static class JavaCastValuePrinter extends JavaPrinter<CastValue>{
		
		public void print(CastValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception {
			writer.write(buffer+"(");
			writer.write("(");
			writer.write(element.getType());
			writer.write(")");
			Value value = element.getValue();
			provider.getPrinter(value).print(value, writer, "", provider);
			writer.write(")");
		}
		
	}
	
	public static class JavaInstanceValuePrinter extends JavaPrinter<InstanceValue>{
		
		public void print(InstanceValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception {
			writer.write(buffer+"new ");
			writer.write(element.getToken());
			writer.write("(");
			Iterator<Value> parameters = element.getParameters().iterator();
			while(parameters.hasNext()){
				Value value = parameters.next();
				provider.getPrinter(value).print(value, writer, "", provider);
				if(parameters.hasNext()) writer.write(", ");
			}
			writer.write(")");
		}
		
	}
	
	public static class JavaLambdaValuePrinter extends JavaPrinter<LambdaValue>{
		
		public void print(LambdaValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception {
			writer.write(buffer+"(");
			Iterator<LambdaParameter> parameters = element.getParameters().iterator();
			while(parameters.hasNext()){
				LambdaParameter parameter = parameters.next();
				if(parameter.getType() != null){
					writer.write(parameter.getType());
					writer.write(" ");
				}
				writer.write(parameter.getName());
				if(parameters.hasNext()) writer.write(", ");
			}
			writer.write(") -> ");
			if(element.getOperations().size() == 1){
				JavaElement value = element.getOperations().get(0);
				provider.getPrinter(value).print(value, writer, "", provider);
			}else{
				writer.write("{\n");
				Iterator<JavaElement> operations = element.getOperations().iterator();
				while(operations.hasNext()){
					JavaElement operation = operations.next();
					provider.getPrinter(operation).print(operation, writer, buffer+"\t", provider);
					if(operation instanceof Value) writer.write(";");
					writer.write("\n");
				}
				writer.write("}");
			}
		}
		
	}
	
	public static class JavaMethodCallValuePrinter extends JavaPrinter<MethodCallValue>{
		
		public void print(MethodCallValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception {
			writer.write(buffer);
			Value base = element.base();
			if(base != null){
				provider.getPrinter(base).print(base, writer, "", provider);
				writer.write(".");
			}
			if(element.getGeneric() != null){
				writer.write("<");
				writer.write(element.getGeneric());
				writer.write(">");
			}
			writer.write(element.getToken());
			writer.write("(");
			Iterator<Value> parameters = element.getParameters().iterator();
			while(parameters.hasNext()){
				Value value = parameters.next();
				provider.getPrinter(value).print(value, writer, "", provider);
				if(parameters.hasNext()) writer.write(", ");
			}
			writer.write(")");
		}
		
	}
	
	public static class JavaModifierValuePrinter extends JavaPrinter<ModifierValue>{
		
		public void print(ModifierValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer+element.getModifier());
			Value value = element.getValue();
			provider.getPrinter(value).print(value, writer, "", provider);
		}
		
	}
	
	public static class JavaVariableAccessValuePrinter extends JavaPrinter<VariableAccessValue>{
		
		public void print(VariableAccessValue element, Writer writer, String buffer, JavaPrintProvider provider) throws Exception{
			writer.write(buffer);
			Value value = element.base();
			provider.getPrinter(value).print(value, writer, "", provider);
			writer.write(".");
			writer.write(element.getVariable());
		}
		
	}

}
